print("import: 'pyslet'")
import pyslet

print("import: 'pyslet.http'")
import pyslet.http

print("import: 'pyslet.odata2'")
import pyslet.odata2

print("import: 'pyslet.qtiv1'")
import pyslet.qtiv1

print("import: 'pyslet.qtiv2'")
import pyslet.qtiv2

print("import: 'pyslet.xml20081126'")
import pyslet.xml20081126

